/**********************************************************************
 * Copyright (c) 2007 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.smlif.internal.common;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.StringTokenizer;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.cosmos.rm.smlif.SMLPlugin;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;

/**
 * Provides common utility methods that used by multiple classes
 * under different packages.
 * 
 * @author Ali Mehregani
 */
public class SMLCommonUtil
{	
	public static void openErrorWithDetail(Shell parent, String title, String message, String cause)
	{
		
		class CustomErrorDialog extends ErrorDialog
		{
			private String cause;
			private static final int LIST_ITEM_COUNT = 7;
			private org.eclipse.swt.widgets.List list;
			private boolean isListDisplayed;
			private Clipboard clipboard;
			
			public CustomErrorDialog(Shell parentShell, int severity, String dialogTitle, String message, String cause) 
			{
				super (parentShell, dialogTitle, message, new Status(severity, SMLPlugin.PLUGIN_ID, severity, "", new Throwable(cause)), IStatus.OK | IStatus.INFO | IStatus.WARNING | IStatus.ERROR);
				this.cause = cause;
				isListDisplayed = false;
			}
			
			public void openErrorDialog()
			{
				this.open();
			}			
			
		    			
		    protected org.eclipse.swt.widgets.List createDropDownList(Composite parent) 
		    {
		    	if (isListDisplayed)
		    	{
		    		isListDisplayed = false;		    		
		    		list.dispose();
		    		return list;
		    	}
		    	
		    	isListDisplayed = true;
		    	
		    	// create the list
		    	list = new org.eclipse.swt.widgets.List(parent, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);

		    	// fill the list
		    	if (cause != null)
		    	{
			    	StringTokenizer st = new StringTokenizer (cause, SMLConstants.nl);
			    	String indents = "";
			    	boolean addedIndentation = false;
			    	while (st.hasMoreTokens())
			    	{
			    		String currentToken = st.nextToken();
			    		currentToken = currentToken.replace ('\r', ' ');
			    		currentToken = currentToken.trim();
			    		if (!addedIndentation && currentToken.startsWith("at"))
			    		{
			    			indents += "  ";
			    			addedIndentation = true;
			    		}
			    		else if (addedIndentation && !currentToken.startsWith("at"))
			    			addedIndentation = false;
			    		
			    		
			    		list.add(indents + currentToken);
			    	}
		    	}
		    	
		        GridData data = new GridData(GridData.HORIZONTAL_ALIGN_FILL
		                | GridData.GRAB_HORIZONTAL | GridData.VERTICAL_ALIGN_FILL
		                | GridData.GRAB_VERTICAL);
		        data.heightHint = list.getItemHeight() * LIST_ITEM_COUNT;
		        data.horizontalSpan = 2;
		        list.setLayoutData(data);
		        list.setFont(parent.getFont());
		        
		        Menu copyMenu = new Menu(list);
		        MenuItem copyItem = new MenuItem(copyMenu, SWT.NONE);
		        copyItem.addSelectionListener(new SelectionListener() {
		            /*
		             * @see SelectionListener.widgetSelected (SelectionEvent)
		             */
		            public void widgetSelected(SelectionEvent e) {
		                copyToClipboard();
		            }

		            /*
		             * @see SelectionListener.widgetDefaultSelected(SelectionEvent)
		             */
		            public void widgetDefaultSelected(SelectionEvent e) {
		                copyToClipboard();
		            }
		        });
		        
		        copyItem.setText(SMLMessages.commonCopy);
		        list.setMenu(copyMenu);
		    	
		        return list;
		    }
		    
		    private void copyToClipboard() {
		        if (clipboard != null)
		            clipboard.dispose();
		        StringBuffer clipBoardBuffer = new StringBuffer();		       
		        
		        String[] listEntrySelections = list.getSelection();
		        for (int i = 0; i < listEntrySelections.length; i++)
		        	clipBoardBuffer.append(listEntrySelections[i] + SMLConstants.LINE_SEPARATOR);
		        clipboard = new Clipboard(list.getDisplay());
		        clipboard.setContents(new Object[] { clipBoardBuffer.toString() },
		                new Transfer[] { TextTransfer.getInstance() });
		    }


		}
				
		CustomErrorDialog errorDialog = new CustomErrorDialog (parent, IStatus.ERROR, title, message, cause);
		errorDialog.openErrorDialog();
	}
	
	public static void openErrorWithDetail(final String title, final String message, Throwable t)
	{		
		String stackTrace =getExceptionStackTrace(t);
		Throwable exceptionCause = (t.getCause() != null ? t.getCause() : t);
			
		final String cause = exceptionCause.getClass().getName() + SMLConstants.nl + exceptionCause.getMessage() + SMLConstants.nl + stackTrace;
		
		SMLPlugin.getDefault().getWorkbench().getDisplay().syncExec(new Runnable()
		{

			public void run() 
			{
				openErrorWithDetail(SMLPlugin.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell(), title, message, cause);
				
			}});
		
	}
	
	public static String getExceptionStackTrace(Throwable t)
	{
		Throwable throwable = findCause(t);
				
		StringWriter stackTrace = new StringWriter(); 
		throwable.printStackTrace(new PrintWriter(stackTrace));	
		
		return stackTrace.toString();
	}
	
	public static Throwable findCause (Throwable t)
	{
		Throwable cause = null;
		if (t instanceof InvocationTargetException)
			cause = ((InvocationTargetException)t).getTargetException();			
		else if (t instanceof CoreException)
			cause = ((CoreException)t).getStatus().getException();				
		
		if (cause == null)
			return t;
		return findCause(cause);
	}
	
	
	public static int identifyDocumentType (IFile file)
	{
		Exception exception = null;
		try
		{
			return org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil.identifyDocumentType(file);
		} 
		catch (IOException e)
		{
			exception = e;
		} 
		catch (CoreException e)
		{
			exception = e;
		}
		
		if (exception != null)
		{
			SMLCommonUtil.openErrorWithDetail(SMLMessages.errorContentTypeTitle, 
				NLS.bind(SMLMessages.errorContentType, file.getFullPath()), exception);
		}
		
		return ISMLConstants.TYPE_UNKNOWN;
	}
}
