/**********************************************************************
 * Copyright (c) 2007 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.smlif.internal.editor;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.forms.IManagedForm;
import org.eclipse.ui.forms.editor.FormEditor;
import org.eclipse.ui.forms.editor.FormPage;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.Section;

/**
 * An abstract form page that all other form pages are expected to
 * extend.
 * 
 * @author Ali Mehregani
 */
public abstract class AbstractFormPage extends FormPage implements ModifyListener
{	
	public AbstractFormPage(FormEditor editor, String id, String title)
	{
		super(editor, id, title);
	}

	/**
	 * @see org.eclipse.ui.forms.editor.FormPage#createFormContent(org.eclipse.ui.forms.IManagedForm)
	 */
	protected void createFormContent(IManagedForm managedForm)
	{
		ScrolledForm form = managedForm.getForm();		
		form.setText(getTitle());
		Composite parent = form.getBody();
		
		GridLayout gridLayout = new GridLayout(2, true);
		gridLayout.marginWidth = 10;
		gridLayout.horizontalSpacing = 15;		
		parent.setLayout(gridLayout);
		parent.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		Composite leftColumn = createColumn(parent);
		Composite rightColumn = createColumn(parent);
		
		createLeftColumn(leftColumn);
		createRightColumn(rightColumn);				
	}	
	
	/**
	 * Create the left column content of the page
	 * 
	 * @param parent The parent composite
	 */
	protected abstract void createLeftColumn(Composite parent);
	
	
	/**
	 * Create the right column content of the page
	 * 
	 * @param parent The parent composite
	 */
	protected abstract void createRightColumn(Composite parent);


	/**
	 * Creates a composite that corresponds to the a column in the given parent.
	 * @param toolkit 
	 * 
	 * @param parent The parent composite
	 * @return Composite The column composite
	 */
	protected Composite createColumn(Composite parent)
	{
		Composite column = getManagedForm().getToolkit().createComposite(parent);
		
		GridLayout gridLayout = new GridLayout();
		gridLayout.verticalSpacing = 10;
		gridLayout.marginHeight = 10;
		gridLayout.marginWidth = 0;
		column.setLayout(gridLayout);
		column.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		
		return column;
	}
	
	
	/**
	 * Create and return a section
	 * 
	 * @return A section
	 */
	protected Section createSection(Composite parent)
	{
		Section section = getManagedForm().getToolkit().createSection(parent, Section.TITLE_BAR | Section.TWISTIE);
		GridData td = new GridData(SWT.FILL, SWT.FILL, true, true);
		section.setLayoutData(td);
		return section;
	}
	
	
	/**
	 * Create a label and a text field.  The controls will be
	 * layedout horizonally:   <label>  <text>
	 * 
	 * @param parent The parent composite
	 * @param label The label text
	 * @param multiLine Whether the text box should be multi-line
	 * @param initialValue The initial value of the text box
	 * @param modifyListener The modify listener that will be registered with the text box
	 */
	protected void createTextField(Composite parent, String label, boolean multiLine, String initialValue, ModifyListener modifyListener)
	{
		FormToolkit toolkit = getManagedForm().getToolkit();
		toolkit.createLabel(parent, label);
		Text text = toolkit.createText(parent, initialValue, multiLine ? SWT.MULTI & SWT.WRAP: SWT.NONE);
		
		if (multiLine)
		{
			final int WIDTH = 100, HEIGHT = 100;
			text.setSize(WIDTH, HEIGHT);
		}
		
		if (initialValue != null)
			text.setText(initialValue);
		
		if (modifyListener != null)
			text.addModifyListener(modifyListener);		
	}


	/**
	 * Equivalent to createTextField(parent, label, false, initialValue, modifyListener)
	 */
	protected void createTextField(Composite parent, String label, String initialValue, ModifyListener modifyListener)
	{
		createTextField (parent, label, false, initialValue, modifyListener);
	}
	
	
	/**
	 * @see org.eclipse.swt.events.ModifyListener#modifyText(org.eclipse.swt.events.ModifyEvent)
	 */
	public void modifyText(ModifyEvent e)
	{
		if (!isDirty())
			getManagedForm().dirtyStateChanged();
	}
	
	
	/**
	 * Set the appropriate properties of the section
	 * 
	 * @param section The section
	 * @param client The section client
	 * @param title The title of the section
	 */
	protected void makeSectionReady(Section section, Composite client, String title)
	{
		section.setClient(client); 
		section.setText("Section title");
		getManagedForm().getToolkit().createCompositeSeparator(section);
	}
}
