package org.eclipse.cosmos.rm.smlif.internal.editor;


import java.io.IOException;
import java.io.InputStream;

import javax.xml.parsers.ParserConfigurationException;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.cosmos.rm.smlif.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.smlif.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.forms.editor.FormEditor;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * The SML-IF editor part used to create three sub-pages: overview,
 * documents, and sml-if
 * 
 * @author Ali Mehregani
 */
public class SMLIFEditor extends FormEditor implements IResourceChangeListener
{

	public static final int TYPE_IDENTITY = 0;

	/**
	 * The overview page
	 */
	private OverviewPage overviewPage;

	/**
	 * The documents pages
	 */
	private DocumentsPage documentsPage;

	/**
	 * The sml-if page
	 */
	private SMLIFPage smlifPage;

	
	/**
	 * The root DOM node of the SML-IF file selected 
	 */
	private Node rootNode;

	/**
	 * The identity node
	 */
	private Node identityNode;

	
	/**
	 * Indicates whether an error has been reported
	 */
	private boolean reportedError;
	
	
	/**
	 * Create the multi-page editor for SML-IF documents
	 */
	public SMLIFEditor() 
	{
		super();
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
	}
	
	
	/**
	 * @see org.eclipse.ui.part.MultiPageEditorPart#init(org.eclipse.ui.IEditorSite, org.eclipse.ui.IEditorInput)
	 */
	public void init(IEditorSite site, IEditorInput input) throws PartInitException
	{
		if (!(input instanceof IFileEditorInput))
			throw new PartInitException (SMLMessages.errorEditorWrongInput);
			
		super.init(site, input);
	}

	/**
	 * @see org.eclipse.ui.forms.editor.FormEditor#addPages()
	 */
	protected void addPages() 
	{
		overviewPage = new OverviewPage(this);
		documentsPage = new DocumentsPage(this);
		smlifPage = new SMLIFPage(this);
	
		
		try
		{
			addPage(overviewPage, getEditorInput());
			addPage(documentsPage, getEditorInput());
			addPage(smlifPage, getEditorInput());
			
			getSite().setSelectionProvider(null);
			setActivePage(OverviewPage.ID);
		} 
		catch (PartInitException e)
		{
			SMLCommonUtil.openErrorWithDetail(SMLMessages.errorEditorTitle, SMLMessages.errorEditorPages, e);
		}
		
		
	}
	
	
	/**
	 * @see org.eclipse.ui.part.MultiPageEditorPart#dispose()
	 */
	public void dispose() 
	{
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		super.dispose();
	}
	
	
	/**
	 * @see org.eclipse.ui.part.EditorPart#doSave(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public void doSave(IProgressMonitor monitor) 
	{
		smlifPage.doSave(monitor);
	}
	
	
	/**
	 * @see org.eclipse.ui.part.EditorPart#doSaveAs()
	 */
	public void doSaveAs() 
	{
		smlifPage.doSaveAs();
		setInput(smlifPage.getEditorInput());
	}
	

	/**
	 * @see org.eclipse.ui.part.EditorPart#isSaveAsAllowed()
	 */
	public boolean isSaveAsAllowed() 
	{
		return true;
	}
	
	
	/**
	 * Closes all project files on project close.
	 */
	public void resourceChanged(final IResourceChangeEvent event)
	{
		if (event.getType() != IResourceChangeEvent.PRE_CLOSE)
			return;
		
		Runnable closeEditors = new Runnable()
		{
			public void run()
			{
				IWorkbenchPage[] pages = getSite().getWorkbenchWindow().getPages();
				for (int i = 0; i<pages.length; i++)
				{
					if(((IFileEditorInput)getEditorInput()).getFile().getProject().equals(event.getResource()))
					{
						IEditorPart editorPart = pages[i].findEditor(getEditorInput());
						pages[i].closeEditor(editorPart,true);
					}
				}
			}    
		};
		
		Display.getDefault().asyncExec(closeEditors);
	}


	public String retrieveField(int type, String name)
	{
		Node rootDocumentNode = getDocumentNode();
		
		/* The identity type */
		if (type == TYPE_IDENTITY)
		{
			identityNode = identityNode == null ? retrieveChildNode(rootDocumentNode, ISMLConstants.SMLIF_URI, ISMLConstants.IDENTITY_ELEMENT) : identityNode;			
			return retrieveTextNode(retrieveChildNode(identityNode, ISMLConstants.SMLIF_URI, name));
		}
		
		return null;
	}

	/**
	 * Retrieve the text child node of the argument node passed
	 * 
	 * @param node The node
	 * @return The child text of node
	 */
	private String retrieveTextNode(Node node)
	{
		// TODO Auto-generated method stub
		return null;
	}


	/**
	 * Retrive the child node of 'parent' with the uri and localName passed in
	 * 
	 * @param parent The parent node
	 * @param uri The URI
	 * @param localName The local name of the element
	 * @return The element node or null if it cannot be found
	 */
	private Node retrieveChildNode(Node parent, String uri, String localName)
	{
		if (parent == null)
			return null;
		
		NodeList children = rootNode.getChildNodes();		
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			Node child = children.item(i);
			if (uri.equals(child.getNamespaceURI()) && localName.equals(child.getLocalName()))
			{
				return child;				
			}
		}
		
		return null;
	}


	/**
	 * Parse and return the DOM document of the SML-IF document
	 * 
	 * @return The DOM document node of the SML-IF file
	 */
	private Node getDocumentNode()
	{
		if (rootNode != null)
			return rootNode;
		
		if (getEditorInput() == null)
			return null;
		
		IFile selectedFile = ((IFileEditorInput)getEditorInput()).getFile();
		if (selectedFile == null)
			return null;
		
		InputStream is = null;
		Exception exception = null;
		try
		{
			is = selectedFile.getContents();
			Document root = org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil.domParseDocument(is);
			NodeList children = root.getChildNodes();			
			if (children.getLength() != 1 || 
				!ISMLConstants.SMLIF_URI.equals((rootNode = children.item(0)).getNamespaceURI()) ||
				!ISMLConstants.MODEL_ELEMENT.equals(rootNode.getLocalName()))
			{
				rootNode = null;				
				throw new SAXException (SMLMessages.errorWrongRootNode);
			}
						
			return rootNode;
			
		} 
		catch (CoreException e)
		{
			exception = e;
		} 
		catch (ParserConfigurationException e)
		{
			exception = e;
		} 
		catch (SAXException e)
		{
			exception = e;
		} 
		catch (IOException e)
		{
			exception = e;
		}
		finally
		{			
			try
			{
				if (is != null)
					is.close();
			} catch (IOException e)
			{
				/* Ignore the exception */
			}
		}
		
		if (exception != null && !reportedError)
		{
			reportedError = true;
			SMLCommonUtil.openErrorWithDetail(SMLMessages.errorOpeningTitle, 
					SMLMessages.errorOpening, 
					exception);
			setActivePage(SMLIFPage.ID);
		}
		
		return null;
	}
}
