/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Path;
import org.eclipse.cosmos.rm.validation.internal.common.AbstractValidationOutput;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil.RemoteRetrievalException;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder.SMLIFIdentity;
import org.eclipse.osgi.util.NLS;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Implements a custom SAX Default Handler that will delegate ContentHandler
 * calls to IDataBuilder instances.
 * 
 * @author sleeloy
 * @author Ali Mehregani
 */
public class DatastructureBuilderImpl extends DefaultHandler implements IValidator
{
	/**
	 * The event types 
	 */
	private static final byte EVENT_CHARACTER = 0x00;				// character event
	private static final byte EVENT_START_DOCUMENT = 0x01;			// start document
	private static final byte EVENT_END_DOCUMENT = 0x02;			// end documnet
	private static final byte EVENT_START_PREFIX = 0x03;			// start prefix mapping
	private static final byte EVENT_END_PREFIX = 0x04;				// end prefix mapping
	private static final byte EVENT_IGNORABLE_WHITESPACE = 0x05;	// ignorable whitespace
	private static final byte EVENT_SKIPPED_ENTITY = 0x06;			// skipped entity
	private static final byte EVENT_PROCESSING_INSTRUCTIONS = 0x08;	// processing instructions
	
	/**
	 * The SAX parser
	 */
	private SAXParser saxParser;

	/**
	 * List of IDataBuilder instances
	 */
	private IDataBuilder[] dataStructureBuilder;

	/**
	 * The validation logger
	 */
	private AbstractValidationOutput validationLogger;

	/**
	 * Custom xml schmea validator
	 */
	private IXMLValidator xmlValidation;

	/**
	 * The xml file to be parsed
	 */
	private String input;

	/**
	 * Indicates whether the XML file should be validated
	 */
	private boolean validateXML;

	/**
	 * The parsing phase that the parser is in
	 */
	private byte phase = ISMLConstants.START_PHASE;

	/**
	 * The locator used to indicate the column and line number of a problem
	 */
	private Locator locator;
	
	/**
	 * A flag used to indicate when a document element has been hit
	 */
	private boolean documentElementHit;
	
	/**
	 * Indicates whether the input is a set of SML unit document or an SML-IF 
	 * document.  
	 */
	private boolean smlIfInput;
	
	public DatastructureBuilderImpl() 
	{
		super();
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.core.IValidator#initialize(java.util.Map)
	 */
	public void initialize(Map validationAttribute)
	{
		validationLogger = (AbstractValidationOutput) validationAttribute.get(ATTRIBUTE_OUTPUT);
		smlIfInput = !(validationAttribute.get(IValidator.ATTRIBUTE_INPUT_TYPE) instanceof String) || !IValidator.VALUE_SML_UNIT.equals(validationAttribute.get(IValidator.ATTRIBUTE_INPUT_TYPE));		
		input = (String)validationAttribute.get(IValidator.ATTRIBUTE_INSTANCE);
		Object validationClass = (validationAttribute.get(IValidator.ATTRIBUTE_VALDIATION_XML));
		
		// if no validation class then we should do no XML schema validation
		validateXML = true;
		if (validationClass == null)
		{
			validateXML = false;
		} 
		else if (validationClass.equals(IXMLValidator.DEFAULT_VALIDATION_XML_LITERAL))
		{
			// should validate and use the defalut validator
			xmlValidation = null;
		} 
		else
		{
			// should use the custom xml schema validator
			IValidator[] validation;
			try
			{
				validation = ValidationFactory.createValidator(IXMLValidator.class, validationClass);
				xmlValidation = validation.length > 0 ? (IXMLValidator) validation[0] : null;
				xmlValidation.initialize(validationAttribute);
			} 
			catch (InstantiationException e)
			{
				validationLogger.reportError(NLS.bind(SMLMessages.errorInstantiating, validationClass));
			} 
			catch (IllegalAccessException e)
			{
				validationLogger.reportError(NLS.bind(SMLMessages.errorInstantiating, validationClass));
			} 
			catch (ClassNotFoundException e)
			{
				validationLogger.reportError(NLS.bind(SMLMessages.errorClassNotFound, validationClass));
			}

		}
		setSchema((String) validationAttribute.get(ATTRIBUTE_SCHEMA));
	}

	/**
	 * Configures the default handler and associates the schema with the SAX
	 * Parser
	 * 
	 * @param xsd file location of the schema file to validate
	 */
	protected void setSchema(String xsd)
	{
		try
		{
			// if there is no schema then we should not validate
			if (xsd == null)
				validateXML = false;

			SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
			saxParserFactory.setNamespaceAware(true);
			saxParserFactory.setValidating(validateXML);

			saxParserFactory.setFeature(ISMLConstants.SAX_PARSER_FEATURE_VALIDATION, validateXML);
			saxParserFactory.setFeature(ISMLConstants.SAX_PARSER_FEATURE_NAMESPACES, true);
			saxParserFactory.setFeature(ISMLConstants.SAX_PARSER_FEATURE_NAMESPACE_PREFIXES, true);

			SAXParser newSaxParser = saxParserFactory.newSAXParser();

			if (xsd != null)
			{
				newSaxParser.setProperty(ISMLConstants.SAX_PARSER_PROPERTY_SCHEMA_LANGUAGE, "http://www.w3.org/2001/XMLSchema");
				newSaxParser.setProperty(ISMLConstants.SAX_PARSER_PROPERTY_SCHEMA_SOURCE, new FileInputStream(new File(xsd)));
			}
			saxParser = newSaxParser;
		} 
		catch (ParserConfigurationException e)
		{
			fatalError(e);
		} 
		catch (IOException e)
		{
			fatalError(e);
		} 
		catch (SAXNotRecognizedException e)
		{
			// Some parsers cannot validate the xsd like crimson. Therefore turn
			// off validation.
			e.printStackTrace();
			fatalError(e);
		} 
		catch (SAXNotSupportedException e)
		{
			// Some parsers cannot validate the xsd like crimson. Therefore turn
			// off validation.
			e.printStackTrace();
			fatalError(e);
		} 
		catch (SAXException e)
		{
			fatalError(e);
		}
	}

	public void error(SAXParseException exception)
	{
		validationLogger.reportError(exception.getLocalizedMessage() + " \nLine:" + exception.getLineNumber() + " \nColumn:" + exception.getColumnNumber());
	}

	protected void fatalError(Exception exception)
	{
		validationLogger.reportError(exception.getLocalizedMessage());
	}

	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.core.IValidator#validate()
	 */
	public boolean validate()
	{
		Exception exception = null;
		String currentDocument = null;
		try
		{
			retrieveBuilders();
			
			/* The input is an SML if document */
			if (smlIfInput)
			{
				currentDocument = input;
				saxParser.parse(new InputSource(new FileInputStream(new File(currentDocument))), this);
			}
			/* Othewise the input is expected to be a set of comma separated files and directories */
			else
			{
				StringTokenizer inputs = new StringTokenizer(input, ",");
				List phenicList = new ArrayList(); 					
				
				/* Register a dummy base URI */
				SMLIFIdentity identity = (SMLIFIdentity)SMLCommonUtil.retrieveDataStructure(IdentityDataBuilder.ID);
				if (identity != null)
					identity.setBaseURI(IValidationConstants.DUMMY_BASE_URI);
				
				
				/* Start the genic element */
				phase = ISMLConstants.GENICS_PHASE;
				elementNotification(ISMLConstants.SMLIF_URI, ISMLConstants.GENICS_ELEMENT, ISMLConstants.GENICS_ELEMENT, 
						IValidationConstants.EMPTY_ATTRIBUTES, true, false, true);
				
				/* For every input */
				while (inputs.hasMoreTokens())
				{
					currentDocument = inputs.nextToken();
					IResource resource = retrieveResource(currentDocument);
					if (resource == null)
						return false;
					traverseRecursively(phenicList, resource, true, IValidationConstants.EMPTY_STRING);
				}

				/* Terminate the genic element */
				elementNotification(ISMLConstants.SMLIF_URI, ISMLConstants.GENICS_ELEMENT, ISMLConstants.GENICS_ELEMENT, 
						IValidationConstants.EMPTY_ATTRIBUTES, true, false, false);
				
				/* Start the phenic element */
				phase = ISMLConstants.PHENICS_PHASE;
				elementNotification(ISMLConstants.SMLIF_URI, ISMLConstants.PHENICS_ELEMENT, ISMLConstants.GENICS_ELEMENT, 
						IValidationConstants.EMPTY_ATTRIBUTES, false, true, true);
				
				/* For every phenic document */
				for (int i = 0, phenicCount = phenicList.size(); i < phenicCount; i++)
				{
					traverseRecursively(phenicList, (IResource)phenicList.get(i), false, IValidationConstants.EMPTY_STRING);
				}
				
				/* Terminate the phenic element */
				elementNotification(ISMLConstants.SMLIF_URI, ISMLConstants.PHENICS_ELEMENT, ISMLConstants.GENICS_ELEMENT, 
						IValidationConstants.EMPTY_ATTRIBUTES, false, true, false);
			}
			
			return true;
		} 
		catch (IOException e)
		{
			exception = e;
		} 
		catch (SAXException e)
		{			
			exception = e;			
		} 
		catch (CoreException e)
		{
			exception = e;
		} 
		catch (ParserConfigurationException e)
		{
			exception = e;
		}

		if (exception != null)
		{
			validationLogger.reportError(NLS.bind(SMLMessages.errorParsingDocument, currentDocument));
			fatalError(exception);
		}
		
		return false;
	}


	private void traverseRecursively(List phenicList, IResource currentResource, boolean isGenic, String parentFolder) throws CoreException, IOException, SAXException, ParserConfigurationException
	{
		if (currentResource instanceof IContainer)
		{
			IResource[] children = ((IContainer)currentResource).members();
			parentFolder = (parentFolder.length() > 0 ? parentFolder + "/" : IValidationConstants.EMPTY_STRING) + currentResource.getName();
			for (int i = 0; i < children.length; i++)
			{
				traverseRecursively(phenicList, children[i], isGenic, parentFolder);
			}			
		}
		else if (currentResource instanceof IFile)
		{			
			int type = SMLCommonUtil.identifyDocumentType((IFile)currentResource);
			
			switch (type)
			{
				/* genic */
				case ISMLConstants.TYPE_GENIC:
					elementNotification(ISMLConstants.SMLIF_URI, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							IValidationConstants.EMPTY_ATTRIBUTES, false, false, true);
					processData((IFile)currentResource);
					elementNotification(ISMLConstants.SMLIF_URI, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							IValidationConstants.EMPTY_ATTRIBUTES, false, false, false);
					break;
				
				/* phenic */
				case ISMLConstants.TYPE_PHENIC:
					if (isGenic)
					{
						phenicList.add(currentResource);
						return;
					}

					/* The document and the doc info section */
					fireNotifications(ISMLConstants.SMLIF_URI, new String[]{
							ISMLConstants.DOCUMENT_ELEMENT,
							ISMLConstants.DOC_INFO_ELEMENT, 
							ISMLConstants.ALIASES_ELEMENT, 
							ISMLConstants.ALIAS_ELEMENT,
						}, true);
					char[] path = ((parentFolder.length() > 0 ? parentFolder + "/" : IValidationConstants.EMPTY_STRING) + currentResource.getName()).toCharArray();
					characters(path, 0, path.length);
					fireNotifications(ISMLConstants.SMLIF_URI, new String[]{
							ISMLConstants.ALIAS_ELEMENT,
							ISMLConstants.ALIASES_ELEMENT,						
							ISMLConstants.DOC_INFO_ELEMENT					
						}, false);
					
					/* The data element */
					processData((IFile)currentResource);
					
					elementNotification(ISMLConstants.SMLIF_URI, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							ISMLConstants.DOCUMENT_ELEMENT, 
							IValidationConstants.EMPTY_ATTRIBUTES, false, false, false);
					break;
			}
		}
	}

	private void processData(IFile file) throws SAXException, ParserConfigurationException, IOException, CoreException
	{
		InputStream fileInputStream = file.getContents();
		try
		{
			elementNotification(ISMLConstants.SMLIF_URI, 
					ISMLConstants.DATA_ELEMENT, 
					ISMLConstants.DATA_ELEMENT, 
					IValidationConstants.EMPTY_ATTRIBUTES, false, false, true);
			Document data = SMLCommonUtil.domParseDocument(fileInputStream);
			NodeList childNodes = data.getChildNodes();
			if (childNodes != null && childNodes.getLength() > 0)
				walkDocument(childNodes.item(0));
			elementNotification(ISMLConstants.SMLIF_URI, 
					ISMLConstants.DATA_ELEMENT, 
					ISMLConstants.DATA_ELEMENT, 
					IValidationConstants.EMPTY_ATTRIBUTES, false, false, false);
		}
		finally
		{
			if (fileInputStream != null)
				fileInputStream.close();
		}
	}

	private void fireNotifications(String uri, String[] elements, boolean open) throws SAXException
	{
		for (int i = 0; i <elements.length; i++)
		{
			elementNotification(uri, elements[i], elements[i], IValidationConstants.EMPTY_ATTRIBUTES, false, false, open);
		}
	}

	private IResource retrieveResource(String path)
	{
		IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();					
		IResource currentResource = root.getFile(new Path(path));	
		currentResource = currentResource.exists() ? currentResource : root.getFolder(new Path(path));		
		if (!currentResource.exists())
		{
			validationLogger.reportError(NLS.bind(SMLMessages.errorMissingFile, path));
			return null;
		}
		
		return currentResource;
	}
	
	
	private void retrieveBuilders()
	{
		if (dataStructureBuilder == null)
		{
			Collection dataBuilders = DataBuilderRegistry.instance().getDataStructureBuilders();
			dataStructureBuilder = new IDataBuilder[dataBuilders.size()];
			dataBuilders.toArray(dataStructureBuilder);
		}
	}

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		boolean startGenicsPhase = false;
		boolean startPhenicsPhase = false;
		
		if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			if (phase != ISMLConstants.PHENICS_PHASE)
			{
				if ((phase != ISMLConstants.GENICS_PHASE) && (ISMLConstants.GENICS_ELEMENT.equals(localName)))
				{
					phase = ISMLConstants.GENICS_PHASE;
					startGenicsPhase = true;
				} else if (ISMLConstants.PHENICS_ELEMENT.equals(localName))
				{
					phase = ISMLConstants.PHENICS_PHASE;
					startPhenicsPhase = true;
				}
			}
			
			if ((phase == ISMLConstants.GENICS_PHASE || phase == ISMLConstants.PHENICS_PHASE))
			{
				if (ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
				{
					documentElementHit = true;
				}
				else if (documentElementHit && ISMLConstants.LOCATOR_ELEMENT.equals(localName))
				{
					retrieveRemoteDocument(attributes);
				}
			}
		}
		
		// build datatructures
		elementNotification (uri, localName, qName, attributes, startGenicsPhase, startPhenicsPhase, true);
	}

	private void retrieveRemoteDocument(Attributes attributes) throws SAXException
	{
		String href = attributes.getValue(ISMLConstants.XLINK_URI, ISMLConstants.HREF_ATTRIBUTE);
		Node node = null;
		Exception exception = null;
		try
		{
			node = SMLCommonUtil.retrieveRemoteDocument(href);			
			Node dataNode = node.getOwnerDocument().createElementNS(ISMLConstants.SMLIF_URI, ISMLConstants.DATA_ELEMENT);
			dataNode.appendChild(node);
			node = dataNode;						
		} 
		catch (RemoteRetrievalException e)
		{
			exception = e;
		}
		
		if (exception != null || node == null)
		{
			validationLogger.reportError(NLS.bind(SMLMessages.errorMissingDocument, String.valueOf(locator.getLineNumber())));						
			throw new SAXException(exception);
		}
		walkDocument(node);
	}

	private void walkDocument(Node node) throws SAXException
	{
		Attributes atts = toAttributes(node.getAttributes());
		domNotification(node, atts, true);
		
		NodeList children = node.getChildNodes();
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			walkDocument (children.item(i));			
		}
		
		domNotification(node, atts, false);
	}

	private void domNotification(Node node, Attributes atts, boolean start) throws SAXException
	{
		switch (node.getNodeType())
		{
			case Node.ELEMENT_NODE:
				elementNotification(node.getNamespaceURI(), node.getLocalName(), node.getNodeName(), 
						atts, false, false, start);
				break;
			case Node.TEXT_NODE:
				if (!start)
					break;
				String value = node.getNodeValue();
				if (value == null)
					break;				
				characters(value.toCharArray(), 0, value.length());
				break;
		}
	}

	private Attributes toAttributes(final NamedNodeMap attributes)
	{				
		AttributeWrapper attributeWrapper = new AttributeWrapper (attributes);
		return attributeWrapper;
	}

	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		boolean endGenicsPhase = false;
		boolean endPhenicsPhase = false;
		
		if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			if ((phase == ISMLConstants.GENICS_PHASE) && (ISMLConstants.GENICS_ELEMENT.equals(localName)))
			{
				phase = ISMLConstants.GENICS_PHENICS_PHASE;
				endGenicsPhase = true;
			} 
			else if ((phase == ISMLConstants.PHENICS_PHASE) && (ISMLConstants.PHENICS_ELEMENT.equals(localName)))
			{
				phase = ISMLConstants.GENICS_PHENICS_PHASE;
				endPhenicsPhase = true;
			}
			
			
			if ((phase == ISMLConstants.GENICS_PHASE || phase == ISMLConstants.PHENICS_PHASE) && ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
			{
				documentElementHit = false;				
			}
		}
		elementNotification (uri, localName, qName, null, endGenicsPhase, endPhenicsPhase, false);
	}
	
	
	private void elementNotification (String uri, String localName, String qName, Attributes attributes, 
									  boolean genicPhase, boolean phenicPhase, boolean start) throws SAXException
	{
		for (int i = 0; i < dataStructureBuilder.length; i++)
		{
			IDataBuilder dataBuilder = dataStructureBuilder[i];
			if (genicPhase)
			{
				if (start)
					dataBuilder.startGenics();
				else
					dataBuilder.endGenics();
			}
			if (phenicPhase)
			{
				if (start)
					dataBuilder.startPhenics();
				else
					dataBuilder.endPhenics();
			}
			if ((dataBuilder.getPhase() == ISMLConstants.UNKNOWN_PHASE) || (dataBuilder.getPhase() == phase) || (dataBuilder.getPhase() == ISMLConstants.GENICS_PHENICS_PHASE && (phase == ISMLConstants.PHENICS_PHASE || phase == ISMLConstants.GENICS_PHASE)))
			{
				if (start)
					dataBuilder.startElement(uri, localName, qName, attributes);
				else
					dataBuilder.endElement(uri, localName, qName);
			}
		}
		
		if (xmlValidation != null)
		{
			if (start)
				xmlValidation.startElement(uri, localName, qName, attributes);
			else
				xmlValidation.endElement(uri, localName, qName);
		}
	}

	public void characters(char[] ch, int start, int len) throws SAXException
	{
		eventNotification (EVENT_CHARACTER, new Object[] {ch, new Integer(start), new Integer(len)});
	}

	public void endDocument() throws SAXException
	{
		eventNotification (EVENT_END_DOCUMENT, null);
	}

	public void endPrefixMapping(String arg0) throws SAXException
	{
		eventNotification (EVENT_END_PREFIX, new Object[]{arg0});
	}

	public void unparsedEntityDecl(String arg0, String arg1, String arg2, String arg3) throws SAXException
	{
		super.unparsedEntityDecl(arg0, arg1, arg2, arg3);
	}

	public void ignorableWhitespace(char[] arg0, int arg1, int arg2) throws SAXException
	{
		eventNotification (EVENT_IGNORABLE_WHITESPACE, new Object[]{arg0, new Integer(arg1), new Integer(arg2)});
	}

	public void processingInstruction(String arg0, String arg1) throws SAXException
	{
		eventNotification (EVENT_PROCESSING_INSTRUCTIONS, new Object[]{arg0, arg1});
	}

	public void skippedEntity(String arg0) throws SAXException
	{
		eventNotification (EVENT_SKIPPED_ENTITY, new Object[]{arg0});
	}

	public void startDocument() throws SAXException
	{
		eventNotification (EVENT_START_DOCUMENT, null);
	}

	public void startPrefixMapping(String arg0, String arg1) throws SAXException
	{
		eventNotification(EVENT_START_PREFIX, new Object[]{arg0, arg1});
	}
	
	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
		retrieveBuilders();
		for (int i = 0; i < dataStructureBuilder.length; i++)
		{
			dataStructureBuilder[i].setDocumentLocator(locator);
		}
		
		
		if (xmlValidation != null)
		{
			xmlValidation.setDocumentLocator(locator);
		}
	}
	
	private void eventNotification (byte event, Object[] arguments) throws SAXException
	{
		retrieveBuilders();
		for (int i = 0; i < dataStructureBuilder.length; i++)
		{
			if ((dataStructureBuilder[i].getPhase() == ISMLConstants.UNKNOWN_PHASE) || (dataStructureBuilder[i].getPhase() == phase) || (dataStructureBuilder[i].getPhase() == ISMLConstants.GENICS_PHENICS_PHASE && (phase == ISMLConstants.PHENICS_PHASE || phase == ISMLConstants.GENICS_PHASE)))
			{
				handleEvent (dataStructureBuilder[i], event, arguments);				
			}
		}		
		
		if (xmlValidation != null)
		{
			handleEvent (xmlValidation, event, arguments);
		}
	}
	
	private void handleEvent (ContentHandler contentHandler, byte event, Object[] arguments) throws SAXException
	{
		switch (event)
		{
			case EVENT_CHARACTER:
				contentHandler.characters((char[])arguments[0], ((Integer)arguments[1]).intValue(), ((Integer)arguments[2]).intValue());
				break;
			case EVENT_START_DOCUMENT:
				contentHandler.startDocument();
				break;
			case EVENT_END_DOCUMENT:
				contentHandler.endDocument();
				break;
			case EVENT_START_PREFIX:
				contentHandler.startPrefixMapping((String)arguments[0], (String)arguments[1]);
				break;
			case EVENT_END_PREFIX:
				contentHandler.endPrefixMapping((String)arguments[0]);
				break;
			case EVENT_IGNORABLE_WHITESPACE:
				contentHandler.ignorableWhitespace((char[])arguments[0], ((Integer)arguments[1]).intValue(), ((Integer)arguments[2]).intValue());
				break;
			case EVENT_SKIPPED_ENTITY:
				contentHandler.skippedEntity((String)arguments[0]);
				break;
			case EVENT_PROCESSING_INSTRUCTIONS:
				contentHandler.processingInstruction((String)arguments[0], (String)arguments[1]);
				break;			
		}
	}
	
	
	private class AttributeWrapper implements Attributes
	{
		private NamedNodeMap attributes;
		private Map qNameInx;
		private Map uriInx;
		
		public AttributeWrapper(NamedNodeMap attributes)
		{
			this.attributes = attributes;
			qNameInx = new Hashtable();
			uriInx = new Hashtable();
			
			if (attributes == null)
				return;
			for (int i = 0, attributeCount = attributes.getLength(); i < attributeCount; i++)
			{
				Node att = attributes.item(i);
				Integer inx = new Integer(i);
				qNameInx.put(att.getNodeName(), inx);
				String key = att.getNamespaceURI() == null ? IValidationConstants.EMPTY_STRING : att.getNamespaceURI();
				key += att.getLocalName();
				uriInx.put(key, inx);
			}
		}
		
		public int getIndex(String qName)
		{			
			Integer inx = (Integer)qNameInx.get(qName);
			return inx == null ? -1 : inx.intValue();
		}

		public int getIndex(String uri, String localName)
		{				
			Integer inx = (Integer)uriInx.get(uri == null ? localName : uri+localName);
			return inx == null ? -1 : inx.intValue();
		}

		public int getLength()
		{
			return attributes.getLength();
		}

		public String getLocalName(int index)
		{
			return attributes.item(index).getLocalName();
		}

		public String getQName(int index)
		{				
			return attributes.item(index).getNodeName();
		}

		public String getType(int index)
		{
			return null;
		}

		public String getType(String qName)
		{
			return null;
		}

		public String getType(String uri, String localName)
		{				
			return null;
		}

		public String getURI(int index)
		{
			return attributes.item(index).getNamespaceURI();
		}

		public String getValue(int index)
		{
			Node node = index >= 0 && index < attributes.getLength() ? attributes.item(index) : null;
			return node == null ? null : node.getNodeValue();
		}

		public String getValue(String qName)
		{
			Node node = attributes.getNamedItem(qName);
			return node == null ? null : node.getNodeValue();
		}

		public String getValue(String uri, String localName)
		{
			Node node = attributes.getNamedItemNS(uri, localName); 
			return node == null ? null : node.getNodeValue();
		}
		
	}
}
