/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.databuilders;

import java.util.Hashtable;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Abstract class that provides a default implementation of the IDataBuilder interface.
 * 
 * @author sleeloy
 * @author Ali Mehregani
 */
public abstract class AbstractDataBuilder extends DefaultHandler implements IDataBuilder 
{	
	/** 
	 * Stores line and column information
	 */
	protected Locator locator;
	
	/**
	 * The target namespace
	 */
	private String targetNamespace;
	
	/**
	 * Stores the default namespace
	 */
	private String defaultNamespace;
	
	/**
	 * The prefix-namespace mapping
	 * KEY = A {@link String} indicating a prefix
	 * VALUE = A {@link String} indicating a URL
	 */
	private Map prefixMap;
	
	/**
	 * The validity of the structure generated.  Set to true by
	 * default.
	 */
	private boolean validity;
	
	/**
	 * The error message
	 */
	private String errorMessage;
	
	/**
	 * Keeps track of the current phase
	 */
	private byte currentPhase;
	
	/**
	 * Constructor
	 */
	public AbstractDataBuilder()
	{
		setStructureValidity(true);
		targetNamespace = "";
		prefixMap = new Hashtable();
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#endGenics()
	 */
	public void endGenics() 
	{
		// no-op
	}

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#endPhenics()
	 */
	public void endPhenics() 
	{
		// no-op
	}

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#getPhase()
	 */
	public byte getPhase() 
	{
		return ISMLConstants.GENICS_PHENICS_PHASE;
	}

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#startGenics()
	 */
	public void startGenics() 
	{
		currentPhase = ISMLConstants.GENICS_PHASE;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#startPhenics()
	 */
	public void startPhenics() 
	{
		currentPhase = ISMLConstants.PHENICS_PHASE;
	}
	

	
	/**
	 * @return the currentPhase
	 */
	public byte getCurrentPhase()
	{
		return currentPhase;
	}

	/**
	 * This is used to determine the target name space of the document currently
	 * being parsed
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		if (ISMLConstants.SCHEMA_URI.equals(uri) && ISMLConstants.SCHEMA_ELEMENT.equals(localName))
		{
			targetNamespace = attributes.getValue(ISMLConstants.TARGET_NAMESPACE_ATTRIBUTE);
			
			for (int i = 0, attributeCount = attributes.getLength(); i < attributeCount; i++)
			{
				String attributeName = attributes.getQName(i);				
				int prefixInx = attributeName.indexOf(':');		
				String rawAttributeName = prefixInx >= 0 ? attributeName.substring(0, prefixInx).trim() : attributeName;
				if (ISMLConstants.XML_NS_ATTRIBUTE.equals(rawAttributeName))
				{					
					if (prefixInx == -1)
					{
						defaultNamespace = attributes.getValue(i);
					}
					else if (attributeName.length() > prefixInx + 1)
					{						
						prefixMap.put(attributeName.substring(prefixInx + 1), attributes.getValue(i));
					}					
				}
			}
		}
	
	}
	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		if (ISMLConstants.SCHEMA_URI.equals(uri) && ISMLConstants.SCHEMA_ELEMENT.equals(localName))
		{
			targetNamespace = "";
			prefixMap.clear();
		}
	}
	
	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#setDocumentLocator(org.xml.sax.Locator)
	 */
	public void setDocumentLocator(Locator locator) 
	{
		this.locator = locator;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#isStructureValid()
	 */
	public boolean isStructureValid()
	{		
		return validity;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#setStructureValidity(boolean)
	 */
	public void setStructureValidity(boolean validity)
	{
		this.validity = validity;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#getErrorMessage()
	 */
	public String getErrorMessage()
	{
		return errorMessage;
	}	
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#setErrorMessage(java.lang.String)
	 */
	public void setErrorMessage(String error)
	{
		this.errorMessage = error;
	}
		
	/**
	 * Appends to the error message
	 * 
	 * @param error The error, which is appended as a separate line
	 */
	protected void appendToErrorMessage(String error)
	{
		if (errorMessage == null)
			errorMessage = "";
		
		if (errorMessage.length() > 0)
			errorMessage += IValidationConstants.LINE_SEPARATOR;
		
		errorMessage += error;
	}

	/**
	 * @return the prefix
	 */
	public Map getPrefixMap()
	{
		return prefixMap;
	}

	/**
	 * @param prefix the prefix to set
	 */
	public void setPrefixMap(Map prefix)
	{
		this.prefixMap = prefix;
	}

	/**
	 * @return the targetNamespace
	 */
	public String getTargetNamespace()
	{
		return targetNamespace;
	}

	/**
	 * @param targetNamespace the targetNamespace to set
	 */
	public void setTargetNamespace(String targetNamespace)
	{
		this.targetNamespace = targetNamespace;
	}

	/**
	 * @return the defaultNamespace
	 */
	public String getDefaultNamespace()
	{
		return defaultNamespace;
	}

	/**
	 * @param defaultNamespace the defaultNamespace to set
	 */
	public void setDefaultNamespace(String defaultNamespace)
	{
		this.defaultNamespace = defaultNamespace;
	}
	
	
	/**
	 * A helper method used to tokenize a qualified name into its
	 * prefix and local name.  A string array of size 2 is returned, where
	 * String[0] = prefix or empty string if no prefix is found and 
	 * String[1] = local name
	 * 
	 * @param qName The qualified name of an element/attribute
	 * @return The prefix and local name tokens of the qualified name
	 */
	protected String[] tokenizeName(String qName)
	{
		if (qName == null)
			return new String[2];
		
		int inx = qName.indexOf(':');
		String prefix = inx >= 0 ? qName.substring(0, inx) : IValidationConstants.EMPTY_STRING;
		String localName = inx >= 0 && qName.length() > inx + 1 ? qName.substring(inx + 1) : qName;
		
		return new String[]{prefix, localName};
	}
}
