/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.databuilders;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.SMLActivator;
import org.eclipse.cosmos.rm.validation.internal.artifacts.SchematronNode;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
/**
 * This builder will extract the schematron from the genic document and create a
 * map of namespace, element and schematron fragments.
 * 
 * @author sleeloy
 */
public class ElementSchematronCacheBuilder extends AbstractDataBuilder {
	public static final String ID = SMLActivator.PLUGIN_ID + ".SchemaCacheBuilder"; //$NON-NLS-1$
	
	protected StringBuffer buffer;
	protected String name;
	protected String type;
	protected String targetNamespace;
	
	protected Map dataStructure = new HashMap();
	protected Map schemaAttributes;
	protected boolean schematronFound = false;
	protected Map prefixLookup;
	protected String ruleContext;
	protected Map ruleBinding = new HashMap();
	protected boolean aliasFound = false;
	protected boolean aliasesFound = false;
	protected List aliases;
	protected boolean phenic = false;
	
	public void endElement(String uri, String localName, String qName) throws SAXException {
		if (ISMLConstants.PHENICS_ELEMENT.equals(localName)){
				phenic = false;
				return;
		}
		if (buffer != null){
			buffer.append("</").append(qName).append(">");			 //$NON-NLS-1$ //$NON-NLS-2$
			if ((schematronFound) && ((ISMLConstants.SCHEMA_ELEMENT.equals(localName) && ISMLConstants.SCHEMATRON_URI.equals(uri)))){
				Map schematronLookup = (Map)dataStructure.get(uri);
				if (schematronLookup == null){
					schematronLookup = new HashMap();
					dataStructure.put(targetNamespace, schematronLookup);
				}
				SchematronNode schematronNode = new SchematronNode(buffer, ruleContext, prefixLookup);
				//associate schematron to element name if found
				if (name != null)
					schematronLookup.put(name, schematronNode);
				//associate schematron to element type if found
				if (type != null)
					schematronLookup.put(type+ISMLConstants.TYPEDELIM, schematronNode);
				//associate schematron to aliases if found
				if (aliases != null){
					Iterator iterRules = ruleBinding.keySet().iterator();
					while (iterRules.hasNext()){
						Object ruleAlias = iterRules.next();
						Object documentAlias = ruleBinding.get(ruleAlias);
						if (aliases.contains(documentAlias)){
							schematronLookup.put(ruleAlias, schematronNode);							
						}
					}
				}
				prefixLookup = new HashMap();
				buffer = null;
				ruleContext = null;
				name = null;
				schematronFound = false;
				aliases = null;
			}			
		}
		if (ISMLConstants.SCHEMA_ELEMENT.equals(localName) && ISMLConstants.SCHEMA_URI.equals(uri)){
			schemaAttributes = null;
		}
		else if (ISMLConstants.ALIASES_ELEMENT.equals(localName) && ISMLConstants.SMLIF_URI.equals(uri)){
			aliasesFound = false;
		}
		else if ((aliasFound) && (ISMLConstants.ALIAS_ELEMENT.equals(localName) && ISMLConstants.SMLIF_URI.equals(uri))){
			aliasFound = false;
		}
		else if (ISMLConstants.ELEMENT_ELEMENT.equals(localName) && ISMLConstants.SCHEMA_URI.equals(uri)){
			name = null;
		}
		else if (ISMLConstants.COMPLEXTYPE_ELEMENT.equals(localName) && ISMLConstants.SCHEMA_URI.equals(uri)){
			type = null;
		}

	}

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException 
	{
		
		if (ISMLConstants.PHENICS_ELEMENT.equals(localName))
		{
			phenic = true;
			return;
		}
		
		if (ISMLConstants.SCHEMA_URI.equals(uri))
		{
			if (ISMLConstants.SCHEMA_ELEMENT.equals(localName)){
				schematronFound = false;			
				targetNamespace = attributes.getValue(ISMLConstants.TARGET_NAMESPACE_ATTRIBUTE);
				//need to keep track of namespace to apply to the sch:schema since the schematron will be extracted in its own document
				schemaAttributes = new HashMap();
				for (int x = 0; x < attributes.getLength(); x++){
					String key = attributes.getQName(x);
					if (key.startsWith(ISMLConstants.XML_NS_ATTRIBUTE)){
						schemaAttributes.put(attributes.getQName(x), attributes.getValue(x));
					}
				}
			}
			else if (ISMLConstants.ELEMENT_ELEMENT.equals(localName))
			{
				name = attributes.getValue(ISMLConstants.NAME_ATTRIBUTE);			
			}
			else if (ISMLConstants.COMPLEXTYPE_ELEMENT.equals(localName))
			{
				type = attributes.getValue(ISMLConstants.NAME_ATTRIBUTE);
			}			
		}
		else if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			if (ISMLConstants.RULEBINDING_ELEMENT.equals(localName))
			{
				ruleBinding.put(attributes.getValue(ISMLConstants.DOCUMENTALIAS_ATTRIBUTE), attributes.getValue(ISMLConstants.RULEALIAS_ATTRIBUTE));
			}
			else if (ISMLConstants.ALIASES_ELEMENT.equals(localName))
			{
				aliasesFound = true;
				aliases = new ArrayList();
			}
			else if ((aliasesFound) && (ISMLConstants.ALIAS_ELEMENT.equals(localName)))
			{
				aliasFound = true;
			}
		}
		else if (ISMLConstants.SCHEMATRON_URI.equals(uri))
		{
			if ((ISMLConstants.SCHEMA_ELEMENT.equals(localName)))
			{
				schematronFound = true;		
				buffer = new StringBuffer();
			}
		}
		
		if (buffer != null){
			if (uri.equals(ISMLConstants.SCHEMATRON_URI)){
				if(localName.equals(ISMLConstants.RULE_ELEMENT)){
					//change rule context if it starts with '.'
					buffer.append("<").append(qName).append(printRuleAttributes(attributes)).append(">"); //$NON-NLS-1$ //$NON-NLS-2$				
				}
				else{
					if (localName.equals(ISMLConstants.NS_ELEMENT)){
						if (prefixLookup == null){
							prefixLookup = new HashMap();
						}
						prefixLookup.put(attributes.getValue(ISMLConstants.PREFIX_ATTRIBUTE), attributes.getValue(ISMLConstants.URI_ATTRIBUTE).trim());
					}
					buffer.append("<").append(qName).append(printReplaceSMLFNAttributes(attributes)).append(">");							 //$NON-NLS-1$ //$NON-NLS-2$
				}
			}
			else if (ISMLConstants.ELEMENT_ELEMENT.equals(localName) && ISMLConstants.SCHEMA_URI.equals(uri)){
				buffer.append("<").append(qName).append(printSchNamespaceAttributes(attributes)).append(">");							 //$NON-NLS-1$ //$NON-NLS-2$
				
			}
			else{
				buffer.append("<").append(qName).append(printAttributes(attributes)).append(">"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
	}
	
	protected StringBuffer printRuleAttributes(Attributes attributes){
		StringBuffer buffer = new StringBuffer();
		for (int x = 0; x<attributes.getLength(); x++){
			String attributeValue = attributes.getValue(x);
			if (attributes.getLocalName(x).equals(ISMLConstants.CONTEXT_ATTRIBUTE)){
				//replace . context with the first node of the document being transformed.
				//This is necessary in the schematron validation phase
				if (attributeValue.startsWith(".")){
					attributeValue = attributeValue.substring(1);
				}
				ruleContext = attributeValue;
			}
			buffer.append(" ").append(attributes.getQName(x)).append("=\"").append(ISMLConstants.CONTEXTPARAM).append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return buffer;
		
	}	
	
	protected StringBuffer printReplaceSMLFNAttributes(Attributes attributes){
		StringBuffer buffer = new StringBuffer();
		for (int x = 0; x<attributes.getLength(); x++){
			if ((attributes.getLocalName(x).equals(ISMLConstants.URI_ATTRIBUTE)) && (attributes.getValue(x).equals(ISMLConstants.SML_FN_URI)|| attributes.getValue(x).equals(ISMLConstants.SML_FN_URI2)))
				buffer.append(" ").append(attributes.getQName(x)).append("=\"").append("org.eclipse.cosmos.rm.validation.internal.reference.SMLExtensionFunctions").append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
			else
				buffer.append(" ").append(attributes.getQName(x)).append("=\"").append(attributes.getValue(x)).append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return buffer;
		
	}	
	protected StringBuffer printSchNamespaceAttributes(Attributes attributes){
		StringBuffer buffer = new StringBuffer();
		for (int x = 0; x<attributes.getLength(); x++){
			buffer.append(" ").append(attributes.getQName(x)).append("=\"").append(attributes.getValue(x)).append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			schemaAttributes.remove(attributes.getQName(x));
		}
		Iterator iter = schemaAttributes.keySet().iterator();
		while (iter.hasNext()){
			Object key = iter.next();
			buffer.append(" ").append(key).append("=\"").append(schemaAttributes.get(key)).append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return buffer;
		
	}
	
	
	protected StringBuffer printAttributes(Attributes attributes){
		StringBuffer buffer = new StringBuffer();
		for (int x = 0; x<attributes.getLength(); x++){
			buffer.append(" ").append(attributes.getQName(x)).append("=\"").append(attributes.getValue(x)).append("\"");				 //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}
		return buffer;
		
	}

	public Object getDataStructure() {
		return dataStructure;
	}

	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
	 */
	public void characters(char[] ch, int start, int length) throws SAXException
	{
		if (phenic) return;
		if (buffer != null){
			buffer.append(new String(ch, start, length));			
		}
		if (aliasFound){
			aliases.add(new String(ch, start, length) );
		}
	}

	public byte getPhase() {
		return ISMLConstants.UNKNOWN_PHASE;
	}
	
	
}
